/*
 * $Id: linux.c,v 1.2 2005/11/21 13:46:28 c4chris Exp $
 *
 * Christian Iseli
 * Ludwig Institute for Cancer Research
 * UNIL - BEP
 * CH-1015 Lausanne
 * Switzerland
 *
 */
#include "os.h"

#ifdef USE_LINUX

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <errno.h>

#include "nfswatch.h"
#include "externs.h"

static void
linux_get_def_dev(char **device)
{
	int n, s;
	struct ifreq *ifrp;
	struct ifconf ifc;
	char buf[BUFSIZ];

	/*
	 * Grab a socket.
	 */
	if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		error("socket");
		finish(-1);
	}

	ifc.ifc_buf = buf;
	ifc.ifc_len = sizeof(buf);

	/*
	 * See what devices we've got.
	 */
	if (ioctl(s, SIOCGIFCONF, (char *) &ifc) < 0) {
		error("ioctl: SIOCGIFCONF");
		finish(-1);
	}

	/*
	 * Take the first device we encounter.
	 */
	ifrp = ifc.ifc_req;
	for (n = ifc.ifc_len/sizeof(struct ifreq); n > 0; n--,ifrp++) {
		/*
		 * Skip the loopback interface.
		 */
		if (strcmp(ifrp->ifr_name, "lo") == 0)
			continue;

		*device = strdup(ifrp->ifr_name);
		break;
	}

	(void) close(s);
}

/*
 * setup_linux_dev - set up the network interface tap.
 */
int
setup_linux_dev(char **device, linux_socket_handle_p_t ls)
{
	int			device_id, err;
	struct ifreq		ifr;
	struct sockaddr_ll	sll;
	socklen_t		errlen = sizeof(err);
	struct packet_mreq	mr;

	/*
	 * If the interface device was not specified,
	 * get the default one.
	 */
	if (*device == NULL)
		linux_get_def_dev(device);
	if (*device == NULL) {
		error("linux: couldn't determine default device");
		finish(-1);
	}

	/* Open the packet socket.  */
	if ((ls->s = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) < 0) {
		perror("linux: socket");
		finish(-1);
	}

	/* Determine link type.  */
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, *device, sizeof(ifr.ifr_name));
	if (ioctl(ls->s, SIOCGIFHWADDR, &ifr) == -1) {
		perror("ioctl");
		finish(-1);
	}
	if (ifr.ifr_hwaddr.sa_family != ARPHRD_ETHER) {
		error("linux: no ethernet device");
		finish(-1);
	}
	ls->linktype = DLT_EN10MB;
	ls->offset = 2;

	/* Determine interface ID.  */
	memset(&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, *device, sizeof(ifr.ifr_name));
	if (ioctl(ls->s, SIOCGIFINDEX, &ifr) == -1) {
		perror("ioctl");
		finish(-1);
	}
	device_id = ifr.ifr_ifindex;

	/* Bind to the device.  */
	memset(&sll, 0, sizeof(sll));
	sll.sll_family		= AF_PACKET;
	sll.sll_ifindex		= device_id;
	sll.sll_protocol	= htons(ETH_P_ALL);
	if (bind(ls->s, (struct sockaddr *) &sll, sizeof(sll)) == -1) {
		perror("bind");
		finish(-1);
	}

	/* Check for errors.  */
	if (getsockopt(ls->s, SOL_SOCKET, SO_ERROR, &err, &errlen) == -1) {
		perror("getsockopt");
		finish(-1);
	}
	if (err > 0) {
		fprintf(stderr, "SO_ERROR: %s (%d)\n",
			strerror(err), err);
		finish(-1);
	}

	/* Go into promiscuous mode.  */
	memset(&mr, 0, sizeof(mr));
	mr.mr_ifindex = device_id;
	mr.mr_type    = PACKET_MR_PROMISC;
	if (setsockopt(ls->s, SOL_PACKET,
			PACKET_ADD_MEMBERSHIP, &mr, sizeof(mr)) == -1)
	{
		perror("setsockopt");
		finish(-1);
	}

	/* Grab a big enough buffer.  */
	ls->bufsize = 65536;
	ls->buffer = (u_char *)malloc(ls->bufsize + ls->offset);
	if (ls->buffer == NULL) {
		perror("malloc");
		finish(-1);
	}
	ls->device = *device;

	return(ls->s);
}

int
linux_read_packet(linux_socket_handle_p_t ls)
{
	u_char	*bp;
	int	packet_len;

	/* Receive a single packet from the kernel */
	bp = ls->buffer + ls->offset;
	do {
		struct sockaddr_ll	from;
		socklen_t		fromlen = sizeof(from);
		packet_len = recvfrom(
			ls->s, bp,
			ls->bufsize, 0,
			(struct sockaddr *) &from, &fromlen);
	} while (packet_len == -1 && errno == EINTR);

	/* Check if an error occured */

	if (packet_len == -1) {
		if (errno == EAGAIN)
			return 0;	/* no packet there */
		else {
			perror("recvfrom");
			return -1;
		}
	}

	/* Get the time stamp of the last packet.  */
	if (ioctl(ls->s, SIOCGSTAMP, &ls->ts) == -1) {
		perror("ioctl");
		return -1;
	}
	ls->bp = bp;
	ls->len = packet_len;

	return 1;
}
#endif /* USE_LINUX */
